# Copyright (C) 2020-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information


from django.core.cache import cache
from django.test import modify_settings, override_settings
import pytest

from swh.auth.django.utils import oidc_profile_cache_key, reverse


@pytest.mark.django_db
@override_settings(SWH_AUTH_SESSION_EXPIRED_REDIRECT_VIEW=None)
def test_oidc_session_expired_middleware_missing_setting(client, keycloak_oidc):
    client.login(code="", code_verifier="", redirect_uri="")
    keycloak_oidc.authorization_code.assert_called()

    url = reverse("root")

    with pytest.raises(ValueError, match="setting is mandatory"):
        client.get(url)


@pytest.mark.django_db
@modify_settings(
    MIDDLEWARE={"remove": ["swh.auth.django.middlewares.OIDCSessionExpiredMiddleware"]}
)
def test_oidc_session_expired_middleware_disabled(client, keycloak_oidc):
    # authenticate user

    client.login(code="", code_verifier="", redirect_uri="")
    keycloak_oidc.authorization_code.assert_called()

    url = reverse("root")

    # visit url first to get user from response
    response = client.get(url)
    assert response.status_code == 200

    # simulate OIDC session expiration
    cache.delete(oidc_profile_cache_key(keycloak_oidc, response.wsgi_request.user.id))

    # no redirection when session has expired
    response = client.get(url)
    assert response.status_code == 200


@pytest.mark.django_db
def test_oidc_session_expired_middleware_enabled(client, keycloak_oidc):
    # authenticate user
    client.login(code="", code_verifier="", redirect_uri="")
    keycloak_oidc.authorization_code.assert_called()

    url = reverse("root")

    # visit url first to get user from response
    response = client.get(url)
    assert response.status_code == 200

    # simulate OIDC session expiration
    cache.delete(oidc_profile_cache_key(keycloak_oidc, response.wsgi_request.user.id))

    # should redirect to logout page
    response = client.get(url)
    assert response.status_code == 302
    silent_refresh_url = reverse(
        "logout", query_params={"next_path": url, "remote_user": 1}
    )
    assert response["location"] == silent_refresh_url



# Copyright (C) 2020-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime, timedelta
from typing import Any, Dict, Optional

from django.conf import settings
from django.contrib.auth.models import Group
from django.http import HttpRequest, QueryDict
from django.urls import reverse as django_reverse

from swh.auth.django.models import OIDCUser
from swh.auth.keycloak import ExpiredSignatureError, KeycloakOpenIDConnect


def oidc_user_from_decoded_token(
    decoded_token: Dict[str, Any], client_id: Optional[str] = None
) -> OIDCUser:
    """Create an OIDCUser out of a decoded token

    Args:
        decoded_token: Decoded token Dict
        client_id: Optional client id of the keycloak client instance used to decode
            the token. If not provided, the permissions will be empty.

    Returns:
        The OIDCUser instance

    """
    # compute an integer user identifier for Django User model
    # by concatenating all groups of the UUID4 user identifier
    # generated by Keycloak and converting it from hex to decimal
    user_id = int("".join(decoded_token["sub"].split("-")), 16)

    # create a Django user that will not be saved to database
    user = OIDCUser(
        id=user_id,
        username=decoded_token.get("preferred_username", ""),
        password="",
        first_name=decoded_token.get("given_name", ""),
        last_name=decoded_token.get("family_name", ""),
        email=decoded_token.get("email", ""),
    )

    # process keycloak groups
    group_names = set()
    if "groups" in decoded_token:
        # set is_staff user property based on group membership
        user.is_staff = "/staff" in decoded_token["groups"]
        for group_name in decoded_token["groups"]:
            # remove leading slash added by keycloak to group name
            django_group_name = group_name.lstrip("/")
            # ensure a corresponding django group exist
            Group.objects.get_or_create(name=django_group_name)
            group_names.add(django_group_name)

    realm_access = decoded_token.get("realm_access", {})
    permissions = realm_access.get("roles", [])

    if client_id:
        # extract user permissions if any
        resource_access = decoded_token.get("resource_access", {})
        client_resource_access = resource_access.get(client_id, {})
        permissions += client_resource_access.get("roles", [])

    # set user permissions and filter out default keycloak realm roles
    user.permissions = set(permissions) - {"offline_access", "uma_authorization"}
    # set user groups
    user.group_names = group_names

    # add user sub to custom User proxy model
    user.sub = decoded_token["sub"]

    return user


def oidc_user_from_profile(
    oidc_client: KeycloakOpenIDConnect, oidc_profile: Dict[str, Any]
) -> OIDCUser:
    """Initialize an OIDCUser out of an oidc profile dict.

    Args:
        oidc_client: KeycloakOpenIDConnect used to discuss with keycloak
        oidc_profile: OIDC profile retrieved once connected to keycloak

    Returns:
        OIDCUser instance parsed out of the token received.

    """

    # decode JWT token
    try:
        access_token = oidc_profile["access_token"]
        decoded_token = oidc_client.decode_token(access_token)
    # access token has expired
    except ExpiredSignatureError:
        # get a new access token from authentication provider
        oidc_profile = oidc_client.refresh_token(oidc_profile["refresh_token"])
        # decode access token
        decoded_token = oidc_client.decode_token(oidc_profile["access_token"])

    # create OIDCUser from decoded token
    user = oidc_user_from_decoded_token(decoded_token, client_id=oidc_client.client_id)

    # get authentication init datetime
    auth_datetime = datetime.fromtimestamp(decoded_token["iat"])
    exp_datetime = datetime.fromtimestamp(decoded_token["exp"])

    # compute OIDC tokens expiration date
    oidc_profile["expires_at"] = exp_datetime
    oidc_profile["refresh_expires_at"] = auth_datetime + timedelta(
        seconds=oidc_profile["refresh_expires_in"]
    )

    # add OIDC profile data to custom User proxy model
    for key, val in oidc_profile.items():
        if hasattr(user, key):
            setattr(user, key, val)

    return user


def oidc_profile_cache_key(oidc_client: KeycloakOpenIDConnect, user_id: int) -> str:
    return f"oidc_user_{oidc_client.realm_name}_{oidc_client.client_id}_{user_id}"


def keycloak_oidc_client() -> KeycloakOpenIDConnect:
    """
    Instantiate a KeycloakOpenIDConnect class from the following django settings:

        * SWH_AUTH_SERVER_URL
        * SWH_AUTH_REALM_NAME
        * SWH_AUTH_CLIENT_ID

    Returns:
        An object to ease the interaction with the Keycloak server

    Raises:
        ValueError: at least one mandatory django setting is not set
    """

    server_url = getattr(settings, "SWH_AUTH_SERVER_URL", None)
    realm_name = getattr(settings, "SWH_AUTH_REALM_NAME", None)
    client_id = getattr(settings, "SWH_AUTH_CLIENT_ID", None)

    if server_url is None or realm_name is None or client_id is None:
        raise ValueError(
            "SWH_AUTH_SERVER_URL, SWH_AUTH_REALM_NAME and SWH_AUTH_CLIENT_ID django "
            "settings are mandatory to instantiate KeycloakOpenIDConnect class"
        )

    return KeycloakOpenIDConnect(
        server_url=server_url, realm_name=realm_name, client_id=client_id
    )


def reverse(
    viewname: str,
    url_args: Optional[Dict[str, Any]] = None,
    query_params: Optional[Dict[str, Any]] = None,
    current_app: Optional[str] = None,
    urlconf: Optional[str] = None,
    request: Optional[HttpRequest] = None,
) -> str:
    """An override of django reverse function supporting query parameters.

    Args:
        viewname: the name of the django view from which to compute a url
        url_args: dictionary of url arguments indexed by their names
        query_params: dictionary of query parameters to append to the
            reversed url
        current_app: the name of the django app tighten to the view
        urlconf: url configuration module
        request: build an absolute URI if provided

    Returns:
        str: the url of the requested view with processed arguments and
        query parameters
    """

    if url_args:
        url_args = {k: v for k, v in url_args.items() if v is not None}

    url = django_reverse(
        viewname, urlconf=urlconf, kwargs=url_args, current_app=current_app
    )

    if query_params:
        query_params = {k: v for k, v in query_params.items() if v is not None}

    if query_params and len(query_params) > 0:
        query_dict = QueryDict("", mutable=True)
        for k in sorted(query_params.keys()):
            query_dict[k] = query_params[k]
        url += "?" + query_dict.urlencode(safe="/;:")

    if request is not None:
        url = request.build_absolute_uri(url)

    return url



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.auth.tests.sample_data import CLIENT_ID, REALM_NAME, SERVER_URL

SECRET_KEY = "o+&ayiuk(y^wh4ijz5e=c2$$kjj7g^6r%z+8d*c0lbpfs##k#7"

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "swh.auth.tests.django.app.apptest",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "swh.auth.django.middlewares.OIDCSessionExpiredMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "swh.auth.tests.django.app.apptest.urls"

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "swh-auth-test-db",
    }
}

SESSION_ENGINE = "django.contrib.sessions.backends.cache"

AUTHENTICATION_BACKENDS = [
    "swh.auth.django.backends.OIDCAuthorizationCodePKCEBackend",
]

SWH_AUTH_SERVER_URL = SERVER_URL
SWH_AUTH_REALM_NAME = REALM_NAME
SWH_AUTH_CLIENT_ID = CLIENT_ID
SWH_AUTH_SESSION_EXPIRED_REDIRECT_VIEW = "logout"

REST_FRAMEWORK = {
    "DEFAULT_RENDERER_CLASSES": ("rest_framework.renderers.JSONRenderer",),
    "DEFAULT_AUTHENTICATION_CLASSES": [
        "rest_framework.authentication.SessionAuthentication",
        "swh.auth.django.backends.OIDCBearerTokenAuthentication",
    ],
}



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# WARNING: do not import unnecessary things here to keep cli startup time under
# control

import sys

import click
from click.core import Context

from swh.core.cli import swh as swh_cli_group

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@swh_cli_group.group(name="auth", context_settings=CONTEXT_SETTINGS)
@click.option(
    "--oidc-server-url",
    "oidc_server_url",
    default="https://auth.softwareheritage.org/auth/",
    help=(
        "URL of OpenID Connect server (default to "
        '"https://auth.softwareheritage.org/auth/")'
    ),
)
@click.option(
    "--realm-name",
    "realm_name",
    default="SoftwareHeritage",
    help=(
        "Name of the OpenID Connect authentication realm "
        '(default to "SoftwareHeritage")'
    ),
)
@click.option(
    "--client-id",
    "client_id",
    default="swh-web",
    help=("OpenID Connect client identifier in the realm " '(default to "swh-web")'),
)
@click.pass_context
def auth(ctx: Context, oidc_server_url: str, realm_name: str, client_id: str):
    """
    Software Heritage Authentication tools.

    This CLI eases the retrieval of a bearer token to authenticate
    a user querying Software Heritage Web APIs.
    """
    from swh.auth.keycloak import KeycloakOpenIDConnect

    ctx.ensure_object(dict)
    ctx.obj["oidc_client"] = KeycloakOpenIDConnect(
        oidc_server_url, realm_name, client_id
    )


@auth.command("generate-token")
@click.argument("username")
@click.pass_context
def generate_token(ctx: Context, username: str):
    """
    Generate a new bearer token for a Web API authentication.

    Login with USERNAME, create a new OpenID Connect session and get
    bearer token.

    Users will be prompted for their password, then the token will be printed
    to standard output.

    The created OpenID Connect session is an offline one so the provided
    token has a much longer expiration time than classical OIDC
    sessions (usually several dozens of days).
    """
    from getpass import getpass

    from swh.auth.keycloak import KeycloakError, keycloak_error_message

    password = getpass()

    try:
        oidc_info = ctx.obj["oidc_client"].login(
            username, password, scope="openid offline_access"
        )
        print(oidc_info["refresh_token"])
    except KeycloakError as ke:
        print(keycloak_error_message(ke))
        sys.exit(1)


@auth.command("revoke-token")
@click.argument("token")
@click.pass_context
def revoke_token(ctx: Context, token: str):
    """
    Revoke a bearer token used for a Web API authentication.

    Use TOKEN to logout from an offline OpenID Connect session.

    The token is definitely revoked after that operation.
    """
    from swh.auth.keycloak import KeycloakError, keycloak_error_message

    try:
        ctx.obj["oidc_client"].logout(token)
        print("Token successfully revoked.")
    except KeycloakError as ke:
        print(keycloak_error_message(ke))
        sys.exit(1)



# Copyright (C) 2020-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Any, Dict, cast
import uuid

from django.contrib.auth import authenticate, login, logout
from django.core.cache import cache
from django.http import HttpRequest
from django.http.response import (
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseRedirect,
    HttpResponseServerError,
)
from django.urls import re_path as url

from swh.auth.django.models import OIDCUser
from swh.auth.django.utils import keycloak_oidc_client, oidc_profile_cache_key, reverse
from swh.auth.keycloak import KeycloakError, keycloak_error_message
from swh.auth.utils import gen_oidc_pkce_codes


def oidc_login_view(request: HttpRequest, redirect_uri: str, scope: str = "openid"):
    """
    Helper view function that initiates a login process using OIDC authorization
    code flow with PKCE.

    OIDC session scope can be modified using the dedicated parameter.
    """
    # generate a CSRF token
    state = str(uuid.uuid4())

    code_verifier, code_challenge = gen_oidc_pkce_codes()

    request.session["login_data"] = {
        "code_verifier": code_verifier,
        "state": state,
        "redirect_uri": redirect_uri,
        "next_path": request.GET.get("next_path", ""),
    }

    authorization_url_params = {
        "state": state,
        "code_challenge": code_challenge,
        "code_challenge_method": "S256",
        "scope": scope,
    }

    try:
        oidc_client = keycloak_oidc_client()
        authorization_url = oidc_client.authorization_url(
            redirect_uri, **authorization_url_params
        )
    except KeycloakError as ke:
        return HttpResponseServerError(keycloak_error_message(ke))

    return HttpResponseRedirect(authorization_url)


def get_oidc_login_data(request: HttpRequest) -> Dict[str, Any]:
    """
    Check and get login data stored in django session.
    """
    if "login_data" not in request.session:
        raise Exception("Login process has not been initialized.")

    login_data = request.session["login_data"]

    if "code" not in request.GET or "state" not in request.GET:
        raise ValueError("Missing query parameters for authentication.")

    # get CSRF token returned by OIDC server
    state = request.GET["state"]

    if state != login_data["state"]:
        raise ValueError("Wrong CSRF token, aborting login process.")

    return login_data


def oidc_login(request: HttpRequest) -> HttpResponse:
    """
    Django view to initiate login process using OpenID Connect authorization
    code flow with PKCE.
    """

    redirect_uri = reverse("oidc-login-complete", request=request)

    return oidc_login_view(request, redirect_uri=redirect_uri)


def oidc_login_complete(request: HttpRequest) -> HttpResponse:
    """
    Django view to finalize login process using OpenID Connect authorization
    code flow with PKCE.
    """
    if "error" in request.GET:
        return HttpResponseServerError(request.GET["error"])

    try:
        login_data = get_oidc_login_data(request)
    except ValueError as ve:
        return HttpResponseBadRequest(str(ve))
    except Exception as e:
        return HttpResponseServerError(str(e))

    next_path = login_data["next_path"] or request.build_absolute_uri("/")

    user = authenticate(
        request=request,
        code=request.GET["code"],
        code_verifier=login_data["code_verifier"],
        redirect_uri=login_data["redirect_uri"],
    )

    if user is None:
        return HttpResponseServerError("User authentication failed.")

    login(request, user)

    return HttpResponseRedirect(next_path)


def oidc_logout(request: HttpRequest) -> HttpResponse:
    """
    Django view to logout using OpenID Connect.
    """
    user = request.user
    logout(request)
    if hasattr(user, "refresh_token"):
        user = cast(OIDCUser, user)
        refresh_token = cast(str, user.refresh_token)
        try:
            # end OpenID Connect session
            oidc_client = keycloak_oidc_client()
            oidc_client.logout(refresh_token)
        except KeycloakError as ke:
            return HttpResponseServerError(keycloak_error_message(ke))
        # remove user data from cache
        cache.delete(oidc_profile_cache_key(oidc_client, user.id))

    return HttpResponseRedirect(request.GET.get("next_path", "/"))


urlpatterns = [
    url(r"^oidc/login/$", oidc_login, name="oidc-login"),
    url(r"^oidc/login-complete/$", oidc_login_complete, name="oidc-login-complete"),
    url(r"^oidc/logout/$", oidc_logout, name="oidc-logout"),
]



# Copyright (C) 2021 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest
from rest_framework.test import APIClient, APIRequestFactory


# Alias rf fixture from pytest-django
@pytest.fixture
def request_factory(rf):
    return rf


# Fixture to get test client from Django REST Framework
@pytest.fixture(scope="module")
def api_client():
    return APIClient()


# Fixture to get API request factory from Django REST Framework
@pytest.fixture(scope="module")
def api_request_factory():
    return APIRequestFactory()



# Copyright (C) 2020-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from click.testing import CliRunner
import pytest

from swh.auth.cli import auth
from swh.auth.tests.sample_data import OIDC_PROFILE

runner = CliRunner()


@pytest.fixture()
def keycloak_oidc(keycloak_oidc, mocker):
    def _keycloak_oidc(server_url, realm_name, client_id):
        keycloak_oidc.server_url = server_url
        keycloak_oidc.realm_name = realm_name
        keycloak_oidc.client_id = client_id
        return keycloak_oidc

    keycloak_oidc_client = mocker.patch("swh.auth.keycloak.KeycloakOpenIDConnect")
    keycloak_oidc_client.side_effect = _keycloak_oidc
    return keycloak_oidc


def _run_auth_command(command, keycloak_oidc, input=None):
    server_url = "http://localhost:5080/auth"
    realm_name = "realm-test"
    client_id = "client-test"
    result = runner.invoke(
        auth,
        [
            "--oidc-server-url",
            server_url,
            "--realm-name",
            realm_name,
            "--client-id",
            client_id,
            *command,
        ],
        input=input,
    )
    assert keycloak_oidc.server_url == server_url
    assert keycloak_oidc.realm_name == realm_name
    assert keycloak_oidc.client_id == client_id
    return result


@pytest.fixture
def user_credentials():
    return {"username": "foo", "password": "bar"}


def test_auth_generate_token_ok(keycloak_oidc, mocker, user_credentials):
    mock_getpass = mocker.patch("getpass.getpass")
    mock_getpass.return_value = user_credentials["password"]

    command = ["generate-token", user_credentials["username"]]
    result = _run_auth_command(
        command, keycloak_oidc, input=f"{user_credentials['password']}\n"
    )
    assert result.exit_code == 0
    assert result.output[:-1] == OIDC_PROFILE["refresh_token"]


def test_auth_generate_token_error(keycloak_oidc, mocker, user_credentials):
    keycloak_oidc.set_auth_success(False)
    mock_getpass = mocker.patch("getpass.getpass")
    mock_getpass.return_value = user_credentials["password"]

    command = ["generate-token", user_credentials["username"]]
    result = _run_auth_command(
        command, keycloak_oidc, input=f"{user_credentials['password']}\n"
    )
    assert result.exit_code == 1
    assert result.output[:-1] == "invalid_grant: Invalid user credentials"


def test_auth_remove_token_ok(keycloak_oidc):
    command = ["revoke-token", OIDC_PROFILE["refresh_token"]]
    result = _run_auth_command(command, keycloak_oidc)
    assert result.exit_code == 0
    assert result.output[:-1] == "Token successfully revoked."


def test_auth_remove_token_error(keycloak_oidc):
    keycloak_oidc.set_auth_success(False)
    command = ["revoke-token", OIDC_PROFILE["refresh_token"]]
    result = _run_auth_command(command, keycloak_oidc)
    assert result.exit_code == 1
    assert result.output[:-1] == "invalid_grant: Invalid user credentials"



# Copyright (C) 2020-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
from urllib.parse import urljoin, urlparse
import uuid

from django.contrib.auth.models import AnonymousUser, User
from django.http import QueryDict
import pytest

from swh.auth.django.models import OIDCUser
from swh.auth.django.utils import reverse
from swh.auth.keycloak import KeycloakError
from swh.auth.tests.django.django_asserts import assert_contains
from swh.auth.tests.sample_data import CLIENT_ID


def _check_oidc_login_code_flow_data(
    request, response, keycloak_oidc, redirect_uri, scope="openid"
):
    parsed_url = urlparse(response["location"])

    authorization_url = keycloak_oidc.well_known()["authorization_endpoint"]
    query_dict = QueryDict(parsed_url.query)

    # check redirect url is valid
    assert urljoin(response["location"], parsed_url.path) == authorization_url
    assert "client_id" in query_dict
    assert query_dict["client_id"] == CLIENT_ID
    assert "response_type" in query_dict
    assert query_dict["response_type"] == "code"
    assert "redirect_uri" in query_dict
    assert query_dict["redirect_uri"] == redirect_uri
    assert "code_challenge_method" in query_dict
    assert query_dict["code_challenge_method"] == "S256"
    assert "scope" in query_dict
    assert query_dict["scope"] == scope
    assert "state" in query_dict
    assert "code_challenge" in query_dict

    # check a login_data has been registered in user session
    assert "login_data" in request.session
    login_data = request.session["login_data"]
    assert "code_verifier" in login_data
    assert "state" in login_data
    assert "redirect_uri" in login_data
    assert login_data["redirect_uri"] == query_dict["redirect_uri"]
    return login_data


@pytest.mark.django_db
def test_oidc_login_views_success(client, keycloak_oidc):
    """
    Simulate a successful login authentication with OpenID Connect
    authorization code flow with PKCE.
    """
    # user initiates login process
    login_url = reverse("oidc-login")

    # should redirect to Keycloak authentication page in order
    # for a user to login with its username / password
    response = client.get(login_url)
    assert response.status_code == 302

    request = response.wsgi_request
    assert isinstance(request.user, AnonymousUser)

    login_data = _check_oidc_login_code_flow_data(
        request,
        response,
        keycloak_oidc,
        redirect_uri=reverse("oidc-login-complete", request=request),
    )

    # once a user has identified himself in Keycloak, he is
    # redirected to the 'oidc-login-complete' view to
    # login in Django.

    # generate authorization code / session state in the same
    # manner as Keycloak
    code = f"{str(uuid.uuid4())}.{str(uuid.uuid4())}.{str(uuid.uuid4())}"
    session_state = str(uuid.uuid4())

    login_complete_url = reverse(
        "oidc-login-complete",
        query_params={
            "code": code,
            "state": login_data["state"],
            "session_state": session_state,
        },
    )

    # login process finalization, should redirect to root url by default
    response = client.get(login_complete_url)
    assert response.status_code == 302

    request = response.wsgi_request
    assert response["location"] == request.build_absolute_uri("/")

    # user should be authenticated
    assert isinstance(request.user, OIDCUser)

    # check remote user has not been saved to Django database
    with pytest.raises(User.DoesNotExist):
        User.objects.get(username=request.user.username)


@pytest.mark.django_db
def test_oidc_logout_view_success(client, keycloak_oidc):
    """
    Simulate a successful logout operation with OpenID Connect.
    """
    # login our test user
    client.login(code="", code_verifier="", redirect_uri="")
    keycloak_oidc.authorization_code.assert_called()

    # user initiates logout
    next_path = reverse("root")
    oidc_logout_url = reverse("oidc-logout", query_params={"next_path": next_path})

    # should redirect to logout page
    response = client.get(oidc_logout_url)
    assert response.status_code == 302

    request = response.wsgi_request
    assert response["location"] == next_path

    # should have been logged out in Keycloak
    oidc_profile = keycloak_oidc.login()
    keycloak_oidc.logout.assert_called_with(oidc_profile["refresh_token"])

    # check effective logout in Django
    assert isinstance(request.user, AnonymousUser)


@pytest.mark.django_db
def test_oidc_login_view_failure(client, keycloak_oidc):
    """
    Simulate a failed authentication with OpenID Connect.
    """
    keycloak_oidc.set_auth_success(False)

    # user initiates login process
    login_url = reverse("oidc-login")
    # should render an error page
    response = client.get(login_url)
    assert response.status_code == 500
    request = response.wsgi_request

    # no users should be logged in
    assert isinstance(request.user, AnonymousUser)


# Simulate possible errors with OpenID Connect in the login complete view.


def test_oidc_login_complete_view_no_login_data(client):
    # user initiates login process
    login_url = reverse("oidc-login-complete")
    # should return with error
    response = client.get(login_url)
    assert response.status_code == 500

    assert_contains(
        response, "Login process has not been initialized.", status_code=500
    )


def test_oidc_login_complete_view_missing_parameters(client):
    # simulate login process has been initialized
    session = client.session
    session["login_data"] = {
        "code_verifier": "",
        "state": str(uuid.uuid4()),
        "redirect_uri": "",
        "next_path": "",
    }
    session.save()

    # user initiates login process
    login_url = reverse("oidc-login-complete")

    # should return with error
    response = client.get(login_url)
    assert response.status_code == 400
    request = response.wsgi_request
    assert_contains(
        response, "Missing query parameters for authentication.", status_code=400
    )

    # no user should be logged in
    assert isinstance(request.user, AnonymousUser)


def test_oidc_login_complete_wrong_csrf_token(client, keycloak_oidc):
    # simulate login process has been initialized
    session = client.session
    session["login_data"] = {
        "code_verifier": "",
        "state": str(uuid.uuid4()),
        "redirect_uri": "",
        "next_path": "",
    }
    session.save()

    # user initiates login process
    login_url = reverse(
        "oidc-login-complete", query_params={"code": "some-code", "state": "some-state"}
    )

    # should render an error page
    response = client.get(login_url)
    assert response.status_code == 400
    request = response.wsgi_request
    assert_contains(
        response, "Wrong CSRF token, aborting login process.", status_code=400
    )

    # no user should be logged in
    assert isinstance(request.user, AnonymousUser)


@pytest.mark.django_db
def test_oidc_login_complete_wrong_code_verifier(client, keycloak_oidc):
    keycloak_oidc.set_auth_success(False)

    # simulate login process has been initialized
    session = client.session
    session["login_data"] = {
        "code_verifier": "",
        "state": str(uuid.uuid4()),
        "redirect_uri": "",
        "next_path": "",
    }
    session.save()

    # check authentication error is reported
    login_url = reverse(
        "oidc-login-complete",
        query_params={"code": "some-code", "state": session["login_data"]["state"]},
    )

    # should render an error page
    response = client.get(login_url)
    assert response.status_code == 500

    request = response.wsgi_request
    assert_contains(response, "User authentication failed.", status_code=500)

    # no user should be logged in
    assert isinstance(request.user, AnonymousUser)


@pytest.mark.django_db
def test_oidc_logout_view_failure(client, keycloak_oidc):
    """
    Simulate a failed logout operation with OpenID Connect.
    """
    # login our test user
    client.login(code="", code_verifier="", redirect_uri="")

    error = "unknown_error"
    error_message = json.dumps({"error": error}).encode()
    keycloak_oidc.logout.side_effect = KeycloakError(
        error_message=error_message, response_code=401
    )

    # user initiates logout process
    logout_url = reverse("oidc-logout")

    # should return with error
    response = client.get(logout_url)
    assert response.status_code == 500
    request = response.wsgi_request
    assert_contains(response, error, status_code=500)

    # user should be logged out from Django anyway
    assert isinstance(request.user, AnonymousUser)



# Copyright (C) 2020-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
from typing import Any, Dict, Optional
from urllib.parse import urlencode

# add ExpiredSignatureError alias to avoid leaking jose import
# in swh-auth client code
from jose.jwt import ExpiredSignatureError  # noqa
from keycloak import KeycloakOpenID

# add KeycloakError alias to avoid leaking keycloak import
# in swh-auth client code
from keycloak.exceptions import KeycloakError  # noqa

from swh.core.config import load_from_envvar


class KeycloakOpenIDConnect:
    """
    Wrapper class around python-keycloak to ease the interaction with Keycloak
    for managing authentication and user permissions with OpenID Connect.
    """

    def __init__(
        self,
        server_url: str,
        realm_name: str,
        client_id: str,
        realm_public_key: str = "",
    ):
        """
        Args:
            server_url: URL of the Keycloak server
            realm_name: The realm name
            client_id: The OpenID Connect client identifier
            realm_public_key: The realm public key (will be dynamically
                retrieved if not provided)
        """
        self._keycloak = KeycloakOpenID(
            server_url=server_url,
            client_id=client_id,
            realm_name=realm_name,
        )
        self.server_url = server_url
        self.realm_public_key = realm_public_key

    @property
    def realm_name(self):
        return self._keycloak.realm_name

    @realm_name.setter
    def realm_name(self, value):
        self._keycloak.realm_name = value

    @property
    def client_id(self):
        return self._keycloak.client_id

    @client_id.setter
    def client_id(self, value):
        self._keycloak.client_id = value

    def well_known(self) -> Dict[str, Any]:
        """
        Retrieve the OpenID Connect Well-Known URI registry from Keycloak.

        Returns:
            A dictionary filled with OpenID Connect URIS.
        """
        try:
            return self._keycloak.well_known()
        except AttributeError:
            # python-keycloak < 1.0.0
            return self._keycloak.well_know()

    def authorization_url(self, redirect_uri: str, **extra_params: str) -> str:
        """
        Get OpenID Connect authorization URL to authenticate users.

        Args:
            redirect_uri: URI to redirect to once a user is authenticated
            extra_params: Extra query parameters to add to the
                authorization URL
        """
        auth_url = self._keycloak.auth_url(redirect_uri)
        if extra_params:
            auth_url += "&%s" % urlencode(extra_params)
        return auth_url

    def authorization_code(
        self, code: str, redirect_uri: str, **extra_params: str
    ) -> Dict[str, Any]:
        """
        Get OpenID Connect authentication tokens using Authorization
        Code flow.

        Raises:
            KeycloakError in case of authentication failures

        Args:
            code: Authorization code provided by Keycloak
            redirect_uri: URI to redirect to once a user is authenticated
                (must be the same as the one provided to authorization_url):
            extra_params: Extra parameters to add in the authorization request
                payload.
        """
        return self._keycloak.token(
            grant_type="authorization_code",
            code=code,
            redirect_uri=redirect_uri,
            **extra_params,
        )

    def login(
        self, username: str, password: str, scope: str = "openid", **extra_params: str
    ) -> Dict[str, Any]:
        """
        Get OpenID Connect authentication tokens using Direct Access Grant flow.

        Raises:
            KeycloakError in case of authentication failures

        Args:
            username: an existing username in the realm
            password: password associated to username
            extra_params: Extra parameters to add in the authorization request
                payload.
        """
        return self._keycloak.token(
            grant_type="password",
            scope=scope,
            username=username,
            password=password,
            **extra_params,
        )

    def refresh_token(self, refresh_token: str) -> Dict[str, Any]:
        """
        Request a new access token from Keycloak using a refresh token.

        Args:
            refresh_token: A refresh token provided by Keycloak

        Returns:
            A dictionary filled with tokens info
        """
        return self._keycloak.refresh_token(refresh_token)

    def decode_token(
        self, token: str, options: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        """
        Try to decode a JWT token.

        Args:
            token: A JWT token to decode
            options: Options for jose.jwt.decode

        Returns:
            A dictionary filled with decoded token content
        """
        if not self.realm_public_key:
            realm_public_key = self._keycloak.public_key()
            self.realm_public_key = "-----BEGIN PUBLIC KEY-----\n"
            self.realm_public_key += realm_public_key
            self.realm_public_key += "\n-----END PUBLIC KEY-----"

        return self._keycloak.decode_token(
            token, key=self.realm_public_key, options=options
        )

    def logout(self, refresh_token: str) -> None:
        """
        Logout a user by closing its authenticated session.

        Args:
            refresh_token: A refresh token provided by Keycloak
        """
        self._keycloak.logout(refresh_token)

    def userinfo(self, access_token: str) -> Dict[str, Any]:
        """
        Return user information from its access token.

        Args:
            access_token: An access token provided by Keycloak

        Returns:
            A dictionary fillled with user information
        """
        return self._keycloak.userinfo(access_token)

    @classmethod
    def from_config(cls, **kwargs: Any) -> "KeycloakOpenIDConnect":
        """Instantiate a KeycloakOpenIDConnect class from a configuration dict.

        Args:

            kwargs: configuration dict for the instance, with one keycloak key, whose
                value is a Dict with the following keys:
                - server_url: URL of the Keycloak server
                - realm_name: The realm name
                - client_id: The OpenID Connect client identifier

        Returns:
            the KeycloakOpenIDConnect instance

        """
        cfg = kwargs["keycloak"]
        return cls(
            server_url=cfg["server_url"],
            realm_name=cfg["realm_name"],
            client_id=cfg["client_id"],
        )

    @classmethod
    def from_configfile(cls, **kwargs: Any) -> "KeycloakOpenIDConnect":

        """Instantiate a KeycloakOpenIDConnect class from the configuration loaded from the
        SWH_CONFIG_FILENAME envvar, with potential extra keyword arguments if their
        value is not None.

        Args:
            kwargs: kwargs passed to instantiation call

        Returns:
            the KeycloakOpenIDConnect instance
        """
        config = dict(load_from_envvar()).get("keycloak", {})
        config.update({k: v for k, v in kwargs.items() if v is not None})
        return cls.from_config(keycloak=config)


def keycloak_error_message(keycloak_error: KeycloakError) -> str:
    """Transform a keycloak exception into an error message."""
    try:
        # keycloak error wrapped in a JSON document
        msg_dict = json.loads(keycloak_error.error_message.decode())
        error_msg = msg_dict["error"]
        error_desc = msg_dict.get("error_description")
        if error_desc:
            error_msg = f"{error_msg}: {error_desc}"
        return error_msg
    except Exception:
        # fallback: return error message string
        return keycloak_error.error_message



# Copyright (C) 2020-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime
import hashlib
from typing import Any, Dict, Optional

from django.core.cache import cache
from django.http import HttpRequest
from django.utils import timezone
from rest_framework.authentication import BaseAuthentication
from rest_framework.exceptions import AuthenticationFailed, ValidationError
import sentry_sdk

from swh.auth.django.models import OIDCUser
from swh.auth.django.utils import (
    keycloak_oidc_client,
    oidc_profile_cache_key,
    oidc_user_from_decoded_token,
    oidc_user_from_profile,
)
from swh.auth.keycloak import (
    ExpiredSignatureError,
    KeycloakError,
    KeycloakOpenIDConnect,
    keycloak_error_message,
)


def _update_cached_oidc_profile(
    oidc_client: KeycloakOpenIDConnect, oidc_profile: Dict[str, Any], user: OIDCUser
) -> None:
    """
    Update cached OIDC profile associated to a user if needed: when the profile
    is not stored in cache or when the authentication tokens have changed.

    Args:
        oidc_client: KeycloakOpenID wrapper
        oidc_profile: OIDC profile used to authenticate a user
        user: django model representing the authenticated user
    """
    # put OIDC profile in cache or update it after token renewal
    cache_key = oidc_profile_cache_key(oidc_client, user.id)
    if (
        cache.get(cache_key) is None
        or user.access_token != oidc_profile["access_token"]
    ):
        # set cache key TTL as refresh token expiration time
        assert user.refresh_expires_at
        ttl = int(user.refresh_expires_at.timestamp() - timezone.now().timestamp())

        # save oidc_profile in cache
        cache.set(cache_key, user.oidc_profile, timeout=max(0, ttl))


class OIDCAuthorizationCodePKCEBackend:
    """
    Django authentication backend using Keycloak OpenID Connect authorization
    code flow with PKCE ("Proof Key for Code Exchange").

    To use that backend globally in your django application, proceed as follow:

        * add ``"swh.auth.django.backends.OIDCAuthorizationCodePKCEBackend"``
          to the ``AUTHENTICATION_BACKENDS`` django setting

        * configure Keycloak URL, realm and client by adding
          ``SWH_AUTH_SERVER_URL``, ``SWH_AUTH_REALM_NAME`` and ``SWH_AUTH_CLIENT_ID``
          in django settings

        * add ``swh.auth.django.views.urlpatterns`` to your django application URLs

        * add an HTML link targeting the ``"oidc-login"`` django view in your
          application views

        * once a user is logged in, add an HTML link targeting the ``"oidc-logout"``
          django view in your application views (a ``next_path`` query parameter
          can be used to redirect to a view of choice once the user is logged out)

    """

    def authenticate(
        self, request: HttpRequest, code: str, code_verifier: str, redirect_uri: str
    ) -> Optional[OIDCUser]:

        user = None
        try:
            oidc_client = keycloak_oidc_client()
            # try to authenticate user with OIDC PKCE authorization code flow
            oidc_profile = oidc_client.authorization_code(
                code, redirect_uri, code_verifier=code_verifier
            )

            # create Django user
            user = oidc_user_from_profile(oidc_client, oidc_profile)

            # update cached oidc profile if needed
            _update_cached_oidc_profile(oidc_client, oidc_profile, user)

        except Exception as e:
            sentry_sdk.capture_exception(e)

        return user

    def get_user(self, user_id: int) -> Optional[OIDCUser]:
        # get oidc profile from cache
        oidc_client = keycloak_oidc_client()
        cache_key = oidc_profile_cache_key(oidc_client, user_id)
        oidc_profile = cache.get(cache_key)
        if oidc_profile:
            try:
                user = oidc_user_from_profile(oidc_client, oidc_profile)
                # update cached oidc profile if needed
                _update_cached_oidc_profile(oidc_client, oidc_profile, user)
                # restore auth backend
                setattr(user, "backend", f"{__name__}.{self.__class__.__name__}")
                return user
            except KeycloakError as ke:
                error_msg = keycloak_error_message(ke)
                if error_msg == "invalid_grant: Session not active":
                    # user session no longer active, remove oidc profile from cache
                    cache.delete(cache_key)
                else:
                    sentry_sdk.capture_exception(ke)
                return None
            except Exception as e:
                sentry_sdk.capture_exception(e)
                return None
        else:
            return None


class OIDCBearerTokenAuthentication(BaseAuthentication):
    """
    Django REST Framework authentication backend using bearer tokens for
    Keycloak OpenID Connect.

    It enables to authenticate a Web API user by sending a long-lived
    OpenID Connect refresh token in HTTP Authorization headers.
    Long lived refresh tokens can be generated by opening an OpenID Connect
    session with the following scope: ``openid offline_access``.

    To use that backend globally in your DRF application, proceed as follow:

        * add ``"swh.auth.django.backends.OIDCBearerTokenAuthentication"``
          to the ``REST_FRAMEWORK["DEFAULT_AUTHENTICATION_CLASSES"]``
          django setting.

        * configure Keycloak URL, realm and client by adding
          ``SWH_AUTH_SERVER_URL``, ``SWH_AUTH_REALM_NAME`` and ``SWH_AUTH_CLIENT_ID``
          in django settings

    Users will then be able to perform authenticated Web API calls by sending
    their refresh token in HTTP Authorization headers, for instance:
    ``curl -H "Authorization: Bearer ${TOKEN}" https://...``.

    """

    def authenticate(self, request):
        auth_header = request.META.get("HTTP_AUTHORIZATION")
        if auth_header is None:
            return None

        try:
            auth_type, refresh_token = auth_header.split(" ", 1)
        except ValueError:
            raise AuthenticationFailed("Invalid HTTP authorization header format")

        if auth_type != "Bearer":
            raise AuthenticationFailed(
                (f"Invalid or unsupported HTTP authorization" f" type ({auth_type}).")
            )
        try:

            oidc_client = keycloak_oidc_client()

            # compute a cache key from the token that does not exceed
            # memcached key size limit
            hasher = hashlib.sha1()
            hasher.update(refresh_token.encode("ascii"))
            cache_key = f"api_token_{hasher.hexdigest()}"

            # check if an access token is cached
            access_token = cache.get(cache_key)

            if access_token is not None:
                # attempt to decode access token
                try:
                    decoded_token = oidc_client.decode_token(access_token)
                # access token has expired
                except ExpiredSignatureError:
                    decoded_token = None

            if access_token is None or decoded_token is None:
                # get a new access token from authentication provider
                access_token = oidc_client.refresh_token(refresh_token)["access_token"]
                # decode access token
                decoded_token = oidc_client.decode_token(access_token)
                # compute access token expiration
                exp = datetime.fromtimestamp(decoded_token["exp"])
                ttl = int(exp.timestamp() - timezone.now().timestamp())
                # save access token in cache while it is valid
                cache.set(cache_key, access_token, timeout=max(0, ttl))

            # create Django user
            user = oidc_user_from_decoded_token(decoded_token, oidc_client.client_id)
        except UnicodeEncodeError:
            raise ValidationError("Invalid bearer token")
        except KeycloakError as ke:
            error_msg = keycloak_error_message(ke)
            if error_msg in (
                "invalid_grant: Offline session not active",
                "invalid_grant: Offline user session not found",
            ):
                error_msg = (
                    "Bearer token expired after a long period of inactivity; "
                    "please generate a new one."
                )
            raise AuthenticationFailed(error_msg)
        except Exception as e:
            raise AuthenticationFailed(str(e))

        return user, None
